/**
 * This file Copyright (c) 2015-2018 Magnolia International
 * Ltd.  (http://www.magnolia-cms.com). All rights reserved.
 *
 *
 * This file is dual-licensed under both the Magnolia
 * Network Agreement and the GNU General Public License.
 * You may elect to use one or the other of these licenses.
 *
 * This file is distributed in the hope that it will be
 * useful, but AS-IS and WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE, TITLE, or NONINFRINGEMENT.
 * Redistribution, except as permitted by whichever of the GPL
 * or MNA you select, is prohibited.
 *
 * 1. For the GPL license (GPL), you can redistribute and/or
 * modify this file under the terms of the GNU General
 * Public License, Version 3, as published by the Free Software
 * Foundation.  You should have received a copy of the GNU
 * General Public License, Version 3 along with this program;
 * if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * 2. For the Magnolia Network Agreement (MNA), this file
 * and the accompanying materials are made available under the
 * terms of the MNA which accompanies this distribution, and
 * is available at http://www.magnolia-cms.com/mna.html
 *
 * Any modifications to this file must keep this entire header
 * intact.
 *
 */
package info.magnolia.ui.vaadin.gwt.client.layout.lazylayout.widget;

import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.Node;
import com.google.gwt.dom.client.Style;
import com.google.gwt.user.client.DOM;
import com.vaadin.client.ComputedStyle;

/**
 * Tracks the base sizes of the thumbnails set from the server, applies those to all the thumbnails,
 * calculates the resulting actual size. Also is capable of applying a scaling ratio to the thumbnails.
 * 6.0 change: always scale thumbnails to fit the full width of view; width to height ratio is constant.
 */
public class ThumbnailsSizeKeeper {

    private static final int MAX_WIDTH = 260; // mock-up size + 1px border
    private static final int MAX_HEIGHT = 208; // mock-up size + 1px border

    private int calculatedWidth;

    private int calculatedHeight;

    private int baseWidth;

    private int baseHeight;

    private Element elementsParent;

    private float ratio;

    private int unscaledWidth;

    private int unscaledHeight;
    private int verticalDecorations;
    private int horizontalDecorations;

    ThumbnailsSizeKeeper(Element elementsParent) {
        this.elementsParent = elementsParent;
    }

    public void scale(float ratio) {
        this.ratio = ratio;
        updateCalculatedOffsetSizes();
        doUpdateAllElementsSize(calculatedWidth, calculatedHeight);
    }

    void scaleToWidth(int width) {
        scale((float)(width - unscaledWidth) / (maxWidth() - baseWidth));
    }

    int height() {
        return this.calculatedHeight;
    }

    int width() {
        return this.calculatedWidth;
    }

    /**
     * @return maximum width of tile, close to default size given by mockup, but scaled to fit full width without remainder.
     */
    int maxWidth() {
        int tiles = (int) Math.round((double) this.elementsParent.getOffsetWidth() / MAX_WIDTH);
        return elementsParent.getOffsetWidth() / tiles;
    }

    boolean updateAllElementsSize(int width, int height) {
        if (this.baseHeight != height || this.baseWidth != width) {
            this.baseHeight = height;
            this.baseWidth = width;

            doUpdateAllElementsSize(width, height);
            updateCalculatedOffsetSizes();
            return true;
        }

        return false;
    }

    void applySizeToElement(Element element) {
        int scaledWidth = scaleWidth(baseWidth, ratio);
        doSetElementSize(scaledWidth, (int) Math.round((double) scaledWidth * MAX_HEIGHT / MAX_WIDTH), element);
    }

    private int scaleWidth(int minValue, double ratio) {
        long scaledWidth = Math.round((maxWidth() - minValue) * ratio + minValue);
        return (int) Math.min(scaledWidth, maxWidth());
    }

    private void doUpdateAllElementsSize(int width, int height) {
        Node element = elementsParent.getFirstChildElement();
        while (element != null) {
            doSetElementSize(width, height, Element.as(element));
            element = element.getNextSibling();
        }
    }

    private void doSetElementSize(int width, int height, Element element) {
        final Style style = element.getStyle();
        style.setFontSize(width * 0.75d, Style.Unit.PX); // can be removed with old UI
        style.setWidth(width, Style.Unit.PX);
        style.setHeight(height, Style.Unit.PX);

        addOrRemoveClass(height <= 155, "caption-hidden", element);
    }

    private void addOrRemoveClass(boolean addClass, String className, Element element) {
        if (addClass) {
            element.addClassName(className);
        } else {
            element.removeClassName(className);
        }
    }

    private void updateCalculatedOffsetSizes() {
        ComputedStyle cs;

        int[] padding;
        int[] border;
        int[] margin;

        if (elementsParent.getChildCount() > 0) {
            final Element firstElement = Element.as(elementsParent.getFirstChild());
            cs = new ComputedStyle(firstElement);

            padding = cs.getPadding();
            border = cs.getBorder();
            margin = cs.getMargin();
        } else {
            final Element stub = Element.as(DOM.createDiv());
            stub.addClassName("thumbnail");
            elementsParent.appendChild(stub);
            cs = new ComputedStyle(stub);
            padding = cs.getPadding();
            border = cs.getBorder();
            margin = cs.getMargin();

            elementsParent.removeChild(stub);
        }

        horizontalDecorations = padding[1] + padding[3] + border[1] + border[3] + margin[1] + margin[3];
        verticalDecorations = padding[0] + padding[2] + border[0] + border[2] + margin[0] + margin[2];

        this.unscaledWidth = baseWidth + horizontalDecorations;
        this.unscaledHeight = baseHeight + verticalDecorations;

        this.calculatedWidth = scaleWidth(baseWidth, ratio);
        this.calculatedHeight = (int) Math.round((double) this.calculatedWidth * MAX_HEIGHT / MAX_WIDTH);

    }

    public float getScaleRatio() {
        return ratio;
    }

    public int getUnscaledWidth() {
        return unscaledWidth;
    }

    public int getUnscaledHeight() {
        return unscaledHeight;
    }

    void setElementsParent(Element elementsParent) {
        this.elementsParent = elementsParent;
    }
}
