/**
 * This file Copyright (c) 2019 Magnolia International
 * Ltd.  (http://www.magnolia-cms.com). All rights reserved.
 *
 *
 * This file is dual-licensed under both the Magnolia
 * Network Agreement and the GNU General Public License.
 * You may elect to use one or the other of these licenses.
 *
 * This file is distributed in the hope that it will be
 * useful, but AS-IS and WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE, TITLE, or NONINFRINGEMENT.
 * Redistribution, except as permitted by whichever of the GPL
 * or MNA you select, is prohibited.
 *
 * 1. For the GPL license (GPL), you can redistribute and/or
 * modify this file under the terms of the GNU General
 * Public License, Version 3, as published by the Free Software
 * Foundation.  You should have received a copy of the GNU
 * General Public License, Version 3 along with this program;
 * if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * 2. For the Magnolia Network Agreement (MNA), this file
 * and the accompanying materials are made available under the
 * terms of the MNA which accompanies this distribution, and
 * is available at http://www.magnolia-cms.com/mna.html
 *
 * Any modifications to this file must keep this entire header
 * intact.
 *
 */
package info.magnolia.ui.vaadin.layout;

import static java.util.stream.Collectors.toList;

import info.magnolia.ui.vaadin.gwt.client.layout.lazylayout.connector.ThumbnailLayoutState;
import info.magnolia.ui.vaadin.gwt.client.layout.lazylayout.rpc.LazyLayoutClientRpc;
import info.magnolia.ui.vaadin.gwt.client.layout.lazylayout.rpc.LazyLayoutServerRpc;
import info.magnolia.ui.vaadin.gwt.client.layout.lazylayout.shared.ThumbnailData;
import info.magnolia.ui.vaadin.gwt.shared.Range;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.vaadin.ui.AbstractComponent;
import com.vaadin.ui.IconGenerator;
import com.vaadin.ui.ItemCaptionGenerator;

/**
 * Base implementation of thumbnail gallery view.
 *
 * @param <T>
 *     data type backing up the thumbnails
 */
public abstract class LazyLayout<T> extends AbstractComponent {

    private static Logger log = LoggerFactory.getLogger(LazyThumbnailLayout.class);

    private final List<ElementSelectionListener<T>> selectionListeners = new ArrayList<>();

    private final List<ElementDblClickListener<T>> dblClickListeners = new ArrayList<>();

    private final List<ElementRightClickListener<T>> rightClickListeners = new ArrayList<>();

    private final Set<T> selectedItems = new HashSet<>();

    private DataProviderKeyMapper mapper = new DataProviderKeyMapper();

    public LazyLayout() {
        LazyLayoutServerRpc serverRpc = getRpc(mapper);
        registerRpc(serverRpc);
    }

    void handleSelectionAtIndex(int index, boolean isMultiple) {
        if (isMultiple) {
            getState().selection.toggleMultiSelection(index);
        } else {
            getState().selection.toggleSelection(index);
        }
        updateSelectedItems();
    }

    void fireSelectionChange() {
        this.onElementsSelected(selectedItems);
    }

    private void updateSelectedItems() {
        selectedItems.clear();
        selectedItems.addAll(getState().selection.selectedIndices.stream()
                .map(selectedItemIndex -> fetchItems(Range.between(0, Integer.MAX_VALUE)).get(selectedItemIndex)).collect(toList()));
    }

    protected abstract ItemCaptionGenerator<T> getItemCaptionGenerator();

    protected abstract IconGenerator<T> getItemResourceGenerator();

    protected abstract List<T> fetchItems(Range added);

    private LazyLayoutServerRpc getRpc(DataProviderKeyMapper mapper) {
        return new LazyLayoutServerRpcImpl(mapper, getRpc(), this);
    }

    public LazyLayoutClientRpc getRpc() {
        return getRpcProxy(LazyLayoutClientRpc.class);
    }

    void onElementDoubleClicked(T item) {
        for (final ElementDblClickListener<T> listener : dblClickListeners) {
            listener.onElementDblClicked(item);
        }
    }

    void onElementRightClicked(T item, int clickX, int clickY) {
        for (final ElementRightClickListener<T> listener : rightClickListeners) {
            listener.onElementRightClicked(item, clickX, clickY);
        }
    }

    private void onElementsSelected(Set<T> items) {
        for (final ElementSelectionListener<T> listener : selectionListeners) {
            listener.onElementsSelected(items);
        }
    }

    List<ThumbnailData> fetchElements(Range range) {
        final List<ThumbnailData> elements = new ArrayList<>(range.length());
        for (int i = range.getStart(); i < range.getEnd(); ++i) {
            final T item = mapper.itemAtIndex(i);
            String thumbnailId = mapper.getKey(item);
            setResource(thumbnailId, getItemResourceGenerator().apply(item));
            ThumbnailData thumbnailData = new ThumbnailData(thumbnailId, true);
            thumbnailData.setCaption(getItemCaptionGenerator().apply(item));
            elements.add(thumbnailData);
        }
        return elements;
    }

    private void setElementAmount(int elementAmount) {
        getState().elementAmount = Math.max(elementAmount, 0);
    }

    public void setElementSize(int width, int height) {
        getState().size.height = height;
        getState().size.width = width;
    }

    public void refresh() {
        if (getState(false).elementAmount > 0) {
            getState().resources.clear();
            mapper.clearAll();
        }

        int itemCount = size();
        setElementAmount(itemCount);
        if (getState().offset > itemCount) {
            getState().offset = 0;
        }

        updateSelectionIndices();
        fireSelectionChange();
        getRpc().refresh();
    }

    public abstract int size();

    public void addElementSelectionListener(final ElementSelectionListener<T> listener) {
        if (listener == null) {
            throw new IllegalArgumentException("Selection listener cannot be null!");
        }
        this.selectionListeners.add(listener);
    }

    public void addDoubleClickListener(final ElementDblClickListener<T> listener) {
        if (listener == null) {
            throw new IllegalArgumentException("Double click listener cannot be null!");
        }
        this.dblClickListeners.add(listener);
    }

    public void addRightClickListener(final ElementRightClickListener<T> listener) {
        if (listener == null) {
            throw new IllegalArgumentException("Right click listener cannot be null!");
        }
        this.rightClickListeners.add(listener);
    }

    @Override
    protected ThumbnailLayoutState getState() {
        return (ThumbnailLayoutState) super.getState();
    }

    @Override
    protected ThumbnailLayoutState getState(boolean markAsDirty) {
        return (ThumbnailLayoutState) super.getState(markAsDirty);
    }

    public void setSelectedItem(T selectedItem) {
        if (selectedItem == null) {
            this.getState().selection.selectedIndices.clear();
        } else {
            this.getState().selection.toggleSelection(-1);
            this.selectedItems.clear();
            this.selectedItems.add(selectedItem);
        }

        updateSelectionIndices();
    }

    @Override
    public void beforeClientResponse(boolean initial) {
        super.beforeClientResponse(initial);

        getState().isFirstUpdate &= initial;
    }

    protected void updateSelectionIndices() {
        this.getState().selection.selectedIndices.clear();
        if (!this.selectedItems.isEmpty()) {
            this.selectedItems.stream().filter(mapper.itemToKey.keySet()::contains).forEach(item -> {
                this.getState().selection.toggleMultiSelection(this.mapper.indexOf(item));
            });
        }
    }

    /**
     * Maps item ids, indices and client-side keys to each other.
     * Highly inspired by Vaadin analogous class used in Grid component implementation
     * (introduced in Vaadin 7.4).
     */
    public class DataProviderKeyMapper implements Serializable {

        private final BiMap<Integer, T> indexToItem = HashBiMap.create();

        private final BiMap<T, String> itemToKey = HashBiMap.create();

        private Range activeRange = Range.withLength(0, 0);

        private long rollingIndex = 0;

        private DataProviderKeyMapper() {
        }

        void setActiveRange(Range newActiveRange) {

            final Range[] removed = activeRange.partitionWith(newActiveRange);
            final Range[] added = newActiveRange.partitionWith(activeRange);

            removeActiveElements(removed[0]);
            removeActiveElements(removed[2]);
            addActiveElements(added[0]);
            addActiveElements(added[2]);

            log.debug("Former active: {}, New Active: {}, idx-id: {}, id-key: {}. Removed: {} and {}, Added: {} and {}",
                    activeRange,
                    newActiveRange,
                    indexToItem.size(),
                    itemToKey.size(),
                    removed[0],
                    removed[2],
                    added[0],
                    added[2]);

            activeRange = newActiveRange;
        }

        private void removeActiveElements(final Range deprecated) {
            for (int i = deprecated.getStart(); i < deprecated.getEnd(); i++) {
                final T item = indexToItem.get(i);

                itemToKey.remove(item);
                indexToItem.remove(i);
            }
        }

        private void addActiveElements(Range added) {
            if (added.isEmpty()) {
                return;
            }

            List<T> newItems = fetchItems(added);
            Integer index = added.getStart();
            for (T items : newItems) {
                if (!indexToItem.containsKey(index)) {
                    if (!itemToKey.containsKey(items)) {
                        itemToKey.put(items, nextKey());
                    }

                    indexToItem.forcePut(index, items);
                }
                index++;
            }
        }

        private String nextKey() {
            return String.valueOf(rollingIndex++);
        }

        public String getKey(T item) {
            String key = itemToKey.get(item);
            if (key == null) {
                key = nextKey();
                itemToKey.put(item, key);
            }
            return key;
        }

        public T getItem(String key) throws IllegalStateException {
            T item = itemToKey.inverse().get(key);
            if (item != null) {
                return item;
            } else {
                throw new IllegalStateException("No item for key " + key + " found.");
            }
        }

        public Collection<T> getItems(Collection<String> keys)
                throws IllegalStateException {
            if (keys == null) {
                throw new IllegalArgumentException("keys may not be null");
            }

            final List<T> items = new ArrayList<>(keys.size());
            for (String key : keys) {
                items.add(getItem(key));
            }
            return items;
        }

        T itemAtIndex(int index) {
            return indexToItem.get(index);
        }

        int indexOf(T item) {
            return indexToItem.inverse().get(item);
        }

        void clearAll() {
            indexToItem.clear();
            itemToKey.clear();
            rollingIndex = 0;
            activeRange = Range.withLength(0, 0);
        }
    }

    /**
     * Listener interface for element selection.
     * @param <T> item type.
     */
    public interface ElementSelectionListener<T> {
        void onElementsSelected(Set<T> items);
    }

    /**
     * Listener for element double clicks.
     * @param <T> item type.
     */
    public interface ElementDblClickListener<T> {
        void onElementDblClicked(T item);
    }

    /**
     * Listener for element right clicks.
     * @param <T> item type.
     */
    public interface ElementRightClickListener<T> {
        void onElementRightClicked(T item, int clickX, int clickY);
    }
}
